import re

class ZXColor(object):
    def __init__(self, index, hidden=False):
        self.index = index & 0x0f
        self.hidden = hidden

        if index & 0x80:
            self.hidden = True

    def __hash__(self):
        return hash(self.index)

    def __eq__(self, other):
        return self.index == int(other)

    def __cmp__(self, other):
        return self.index.__cmp__(int(other))

    def __int__(self):
        return self.index

    def __str__(self):
        color_names = ['black', 'blue', 'red', 'magenta', 'green', 'cyan', 'yellow', 'white']
        if self.bright:
            return color_names[self.index & 0x07].upper()
        else:
            return color_names[self.index]

    def _b(self):
        return True if self.index & 0x01 else False

    def _r(self):
        return True if self.index & 0x02 else False

    def _g(self):
        return True if self.index & 0x04 else False

    def _bright(self):
        return True if self.index & 0x08 else False

    b = property(_b)
    r = property(_r)
    g = property(_g)
    bright = property(_bright)


class Palette(object):

    repository = {
        'pulsar':   '00,76,CD,E9,FF,9F:FF,00,00;00,FF,00;00,00,FF',
        'alone':    '00,60,A0,E0,FF,A0:FF,00,00;00,FF,00;00,00,FF',
        'orthodox': '00,76,CD,E9,FF,9F:D0,00,00;00,E4,00;00,00,FF',
        'Mars':     '00,80,C0,E0,FF,C8:FF,00,00;40,C0,00;00,40,C0',
        'Ocean':    '20,80,A0,C0,E0,A8:D0,00,30;00,D0,30;00,00,FF',
        'Grey':     '00,80,C0,E0,FF,C8:49,92,24;49,92,24;49,92,24',
    }

    def __init__(self, palette):
        '''Palette constructor accepts name of known palettes or
           palette string in Unreal Speccy format
           for example: 00,80,C0,E0,FF,C8:FF,00,00;40,C0,00;00,40,C0
        '''
        self.name = palette
        try:
            palette = Palette.repository[palette]
        except KeyError:
            self.name = 'Custom <%s>' % palette

        components = re.split(r'[,;:]', palette)
        if len(components) != 15:
            raise TypeError()
        try:
            values = [int(c, 16) for c in components]
        except TypeError:
            raise TypeError()

        names = (
            'zz', 'zn', 'nn',
            'nb', 'bb', 'zb',
            'r11', 'r12', 'r13',
            'r21', 'r22', 'r23',
            'r31', 'r32', 'r33'
        )

        for name, value in zip(names, values):
            setattr(self, name, value)

        # dual screen colors are symmetrical
        self.nz = self.zn
        self.bn = self.nb
        self.bz = self.zb

        self.norm = self._norm()
        self._colors = {}

        self.hidden_pixel_color = False

    def _norm(self):
        Sr = self.r11 + self.r12 + self.r13
        Sg = self.r21 + self.r22 + self.r23
        Sb = self.r31 + self.r32 + self.r33

        return max(Sr, Sg, Sb)

    def _level2color(self, r, g, b):
        r_ = int((r*self.r11 + g*self.r12 + b*self.r13)/self.norm)
        g_ = int((r*self.r21 + g*self.r22 + b*self.r23)/self.norm)
        b_ = int((r*self.r31 + g*self.r32 + b*self.r33)/self.norm)

        return (r_, g_, b_)

    def _gigacolor(self, idx):
        if self.hidden_pixel_color and (idx[0].hidden or idx[1].hidden):
            return self.hidden_pixel_color
        try:
            return self._colors[idx]
        except KeyError:
            r = g = b = ''
            for el in idx:
                bright = 'b' if el.bright else 'n'

                b += bright if el.b else 'z'
                r += bright if el.r else 'z'
                g += bright if el.g else 'z'

            r = getattr(self, r)
            g = getattr(self, g)
            b = getattr(self, b)

            color = self._level2color(r, g, b)

            # color matrix is symmetrical
            self._colors[idx] = color
            self._colors[(idx[1], idx[0])] = color
            return color

    def _color(self, idx):
        if self.hidden_pixel_color and idx.hidden:
            return self.hidden_pixel_color

        try:
            return self._colors[idx]
        except KeyError:
            one = self.bb if idx.bright else self.nn

            b = one if idx.b else self.zz
            r = one if idx.r else self.zz
            g = one if idx.g else self.zz

            color = self._level2color(r, g, b)

            self._colors[idx] = color

            return color

    def __getitem__(self, key):
        '''Convert Speccy Color Index to real color

           input
            - <color index> for normal mode
            - (<color index>, <color index>) for double screen mode
        '''
        if isinstance(key, tuple):
            return self._gigacolor(key)
        else:
            return self._color(key)

    def __str__(self):
        return self.name
