import itertools
from PIL import Image

from picture import Picture
from color import ZXColor


class Attr(object):
    def __init__(self, attr):
        attr = ord(attr)
        self.paper = ZXColor((attr >> 3) & 0x0f)
        self.ink = ZXColor((attr & 0x07) | (self.paper.index & 0x08))
        self.flash = True if (self.paper != self.ink and attr & 0x80) else False


# http://martyalchin.com/2008/jan/10/simple-plugin-framework/
class _FormatRepository(type):
    def __init__(cls, name, bases, attrs):
        if not hasattr(cls, 'formats'):
            # This branch only executes when processing the mount point itself.
            # So, since this is a new plugin type, not an implementation, this
            # class shouldn't be registered as a plugin. Instead, it sets up a
            # list where plugins can be registered later.
            cls.formats = []
        else:
            # This must be a plugin implementation, which should be registered.
            # Simply appending it to the list is all that's needed to keep
            # track of it later.
            cls.formats.append(cls)

class Format(object):
    __metaclass__ = _FormatRepository


class Bitplan(object):
    def __init__(self, bytes):
        assert len(bytes) == 6144
        self.bytes = bytes

    def screen(self):
        for y in range(192):
            line = 32 * ((y & 0xC0) | ((y << 3) & 0x38) | ((y >> 3) & 0x07))
            for x in range(32):
                pixels = ord(self.bytes[line+x])
                yield (x, y, pixels)

    def sprite(self):
        for y in range(192):
            line = 32 * y
            for x in range(32):
                pixels = ord(self.bytes[line+x])
                yield (x, y, pixels)

class Screen(Format):
    @staticmethod
    def detect(bytes, fn):
        if len(bytes) not in (6912, 6144):
            return False
        return Screen(bytes)

    def __init__(self, bytes):
        self.flashing = False

        self.bitplan = Bitplan(bytes[:6144])
        if len(bytes) == 6912:
            self.attr = bytes[6144:]
        else:
            # black on white
            self.attr = chr(7)*768

    def pixels(self):
        for x, y, pixels in self.bitplan.screen():
            attr = Attr(self.attr[32*(y/8) + x])
            for mask in (0x80, 0x40, 0x20, 0x10, 0x08, 0x04, 0x02, 0x01):
                pixel = pixels & mask
                color = alt_color = attr.ink if pixel else attr.paper

                if attr.flash:
                    alt_color = attr.paper if pixel else attr.ink

                if pixel and attr.ink == attr.paper:
                    color.hidden = alt_color.hidden = True

                yield (color, alt_color)

    def bitmap(self, palette):
        first = []
        second = []

        for color, alt_color in self.pixels():
            if color != alt_color:
                self.flashing = True

            first.append(palette[color])
            second.append(palette[alt_color])

        im1 = Image.new("RGB", (256, 192))
        im1.putdata(first)

        im2 = Image.new("RGB", (256, 192))
        im2.putdata(second)

        if self.flashing:
            return Picture(im1, im2, format='Screen', durations=(50, 50))
        else:
            return Picture(im1, format='Screen')

class GigaScreen(Format):
    @staticmethod
    def detect(bytes, fn):
        if len(bytes) != 2*6912:
            return False
        return GigaScreen(bytes)

    def __init__(self, bytes):
        self.first = Screen(bytes[:6912])
        self.second = Screen(bytes[6912:])

    def pixels(self):
        for pix1, pix2 in itertools.izip(self.first.pixels(), self.second.pixels()):
            yield (pix1[0], pix2[0])

    def bitmap(self, palette):
        im = Image.new("RGB", (256, 192))
        data = [palette[color] for color in self.pixels()]
        im.putdata(data)

        return Picture(im, format='GigaScreen')

class ThreeColor(Format):
    @staticmethod
    def detect(bytes, fn):
        if len(bytes) != 3*6144:
            return False

        if fn.ext.lower() == '.rgb':
            return ThreeColor(bytes, rgb=True)
        else:
            return ThreeColor(bytes)

    def __init__(self, bytes, rgb=False):
        self.rgb = rgb
        if rgb:
            self.r = Bitplan(bytes[:6144])
            self.g = Bitplan(bytes[6144:12288])
            self.b = Bitplan(bytes[12288:])
        else:
            self.b = Bitplan(bytes[:6144])
            self.r = Bitplan(bytes[6144:12288])
            self.g = Bitplan(bytes[12288:])

    def _pixels(self, bitplan):
        for x, y, pixels in bitplan.screen():
            for mask in (0x80, 0x40, 0x20, 0x10, 0x08, 0x04, 0x02, 0x01):
                yield 1 if pixels & mask else 0

    def pixels(self):
        r = self._pixels(self.r)
        g = self._pixels(self.g)
        b = self._pixels(self.b)

        for b2, b1, b0 in itertools.izip(g, r, b):
            idx = (b2 << 2) | (b1 << 1) | b0
            yield ZXColor(idx)

    def bitmap(self, palette):
        im = Image.new("RGB", (256, 192))
        data = [palette[color] for color in self.pixels()]
        im.putdata(data)

        format = '3Color %s' % ('RGB' if self.rgb else 'BRG')
        return Picture(im, format=format)


class MultiGigaScreen(Format):
    @staticmethod
    def detect(bytes, fn):
        if len(bytes) != 36871:
            return False
        if bytes[:3] != 'MGS':
            return False
        if ord(bytes[4]) not in (1, 2, 4, 8):
            return False

        return MultiGigaScreen(bytes)

    def __init__(self, bytes):
        self.char_size = ord(bytes[4])
        self.bitplan = [Bitplan(bytes[7:6151]), Bitplan(bytes[6151:12295])]
        self.attr = [bytes[12295:24583], bytes[24583:]]

        self.border_color = (ord(bytes[5]), ord(bytes[6]))

    def _pixels(self, idx):
        mgs2giga = [0, 1, 9, 2, 10, 3, 11, 4, 12, 5, 13, 6, 14, 7, 15, 8]

        for x, y, pixels in self.bitplan[idx].sprite():
            attr_line = 64*(y & ~(self.char_size-1))

            # MGS 8x1 supports one color for pixel line only in the middle
            # of the screen, the first 8 and the last 8 characters
            # use one color for 8 pixel lines
            if self.char_size == 1 and (x < 8 or x > 24):
                attr_line = 64*(y & 0xf8)

            ink = ord(self.attr[idx][attr_line + x*2])
            paper = ord(self.attr[idx][attr_line + x*2 + 1])

            ink, paper = mgs2giga[ink], mgs2giga[paper]

            for mask in (0x80, 0x40, 0x20, 0x10, 0x08, 0x04, 0x02, 0x01):
                color = ZXColor(ink if pixels & mask else paper)

                if ink == paper and pixels & mask:
                    color.hidden = True

                yield color

    def pixels(self):
        i1 = self._pixels(0)
        i2 = self._pixels(1)

        return itertools.izip(i1, i2)

    def bitmap(self, palette):
        im = Image.new("RGB", (256, 192))
        data = [palette[color] for color in self.pixels()]
        im.putdata(data)

        b1c = ZXColor(self.border_color[0])
        b2c = ZXColor(self.border_color[1])
        format = "MultiGigaScreen 8x%d" % self.char_size

        return Picture(im, format=format, border_color=palette[(b1c, b2c)])


class MultiGiga8xN(Format):
    attr_size = {8: 768, 4: 1536, 2: 3072}

    @classmethod
    def detect(cls, bytes, fn):
        if len(bytes) < 256:
            return False
        if bytes[:3] != 'MGH':
            return False

        char_size = ord(bytes[4])
        if char_size not in (8, 4, 2):
            return False

        if len(bytes) != 256 + 2*6144 + 2*cls.attr_size[char_size]:
            return False

        return MultiGiga8xN(bytes)

    def __init__(self, bytes):
        self.border_color = (ord(bytes[5]), ord(bytes[6]))
        self.char_size = ord(bytes[4])

        bytes = bytes[256:]

        attr_size = self.attr_size[self.char_size]

        self.bitplan = [
            Bitplan(bytes[:6144]),
            Bitplan(bytes[6144:2*6144]),
        ]
        self.attr = [
            bytes[2*6144:2*6144+attr_size],
            bytes[2*6144+attr_size:]
        ]

    def _pixels(self, idx):
        for x, y, pixels in self.bitplan[idx].screen():
            attr = Attr(self.attr[idx][32*(y/self.char_size)+x])
            for mask in (0x80, 0x40, 0x20, 0x10, 0x08, 0x04, 0x02, 0x01):
                pixel = pixels & mask
                color = attr.ink if pixel else attr.paper

                if pixel and attr.ink == attr.paper:
                    color.hidden = True

                yield color

    def pixels(self):
        first = self._pixels(0)
        second = self._pixels(1)
        return itertools.izip(first, second)

    def bitmap(self, palette):
        im = Image.new("RGB", (256, 192))
        data = [palette[color] for color in self.pixels()]
        im.putdata(data)

        b1c = ZXColor(self.border_color[0])
        b2c = ZXColor(self.border_color[1])

        format = 'MultiGiga8x%d' % self.char_size
        return Picture(im, format=format, border_color=palette[(b1c, b2c)])


class MultiGiga8x1(Format):
    @classmethod
    def detect(cls, bytes, fn):
        if len(bytes) != 256 + 2*6144 + 2*3072 + 2*384:
            return False

        if bytes[:3] != 'MGH':
            return False

        if ord(bytes[4]) != 1:
            return False

        return MultiGiga8x1(bytes)

    def __init__(self, bytes):
        self.border_color = (ord(bytes[5]), ord(bytes[6]))

        bytes = bytes[256:]

        self.bitplan = [
            Bitplan(bytes[:6144]),
            Bitplan(bytes[6144:2*6144])
        ]
        self.mgsattr = [
            bytes[2*6144:2*6144+3072],
            bytes[2*6144+3072:2*(6144+3072)]
        ]
        self.gsattr = [
            bytes[2*(6144+3072):2*(6144+3072)+384],
            bytes[2*(6144+3072)+384:]
        ]

    def _pixels(self, idx):
        for x, y, pixels in self.bitplan[idx].screen():
            mgs_attr_line = 16*y
            gs_attr_line = 16*(y/8)

            if x < 8:
                attr = Attr(self.gsattr[idx][gs_attr_line+x])
            elif x < 24:
                attr = Attr(self.mgsattr[idx][mgs_attr_line+x-8])
            else:
                attr = Attr(self.gsattr[idx][gs_attr_line+x-16])

            for mask in (0x80, 0x40, 0x20, 0x10, 0x08, 0x04, 0x02, 0x01):
                pixel = pixels & mask
                color = attr.ink if pixel else attr.paper

                if attr.ink == attr.paper and pixel:
                    color.hidden = True

                yield color

    def pixels(self):
        first = self._pixels(0)
        second = self._pixels(1)

        return itertools.izip(first, second)

    def bitmap(self, palette):
        im = Image.new("RGB", (256, 192))
        data = [palette[color] for color in self.pixels()]
        im.putdata(data)

        b1c = ZXColor(self.border_color[0])
        b2c = ZXColor(self.border_color[1])

        return Picture(im, format='MultiGiga8x1', border_color=palette[(b1c, b2c)])
