from PIL import Image, ImageDraw, ImageFilter
from PIL.GifImagePlugin import getheader, getdata

def o16(i):
    return chr(i&255) + chr(i>>8&255)

# based on
# http://sites.google.com/site/almarklein/files-1/images2gif.py

def getheaderAnim(im):
    """ Animation header. To replace the getheader()[0] """
    bb = "GIF89a"
    bb += o16(im.size[0])
    bb += o16(im.size[1])
    bb += "\x87\x00\x00"
    return bb

def getAppExt(loops=0):
    """ Application extention. Part that secifies amount of loops.
    if loops is 0, if goes on infinitely.
    """
    bb = "\x21\xFF\x0B"  # application extension
    bb += "NETSCAPE2.0"
    bb += "\x03\x01"
    if loops == 0:
        loops = 2**16-1
    bb += o16(loops)
    bb += '\x00'  # end
    return bb

def getGraphicsControlExt(duration=100):
    """ Graphics Control Extension. A sort of header at the start of
    each image. Specifies transparancy and duration. """
    bb = '\x21\xF9\x04'
    bb += '\x08'  # no transparancy
    bb += o16(duration)
    bb += '\x00'  # no transparant color
    bb += '\x00'  # end
    return bb

def write_animated_gif(fp, images, durations):
    """ Given a set of images writes the bytes to the specified stream."""
    im = images[0]

    header = getheaderAnim(im)
    palette = getheader(im)[1]
    appext = getAppExt()

    # write global header
    fp.write(header)
    fp.write(palette)
    fp.write(appext)

    for i, im in enumerate(images):
        graphext = getGraphicsControlExt(durations[i])
        fp.write(graphext)

        for d in getdata(im):
            fp.write(d)

    fp.write(";")  # end gif


class LightBlur(ImageFilter.BuiltinFilter):
    name = "Light Blur"
    filterargs = (3, 3), 4, 0, (
        1, 1, 0,
        1, 1, 0,
        0, 0, 0
        )

def add_border(images, mode, color):
    def _add_border(im, mode):
        # border makes sense only for full-sceen images
        if im.size != (256, 192):
            return im

        # mode could be:
        #  - integer: 0, 1, 2
        #  - string: '0', '1', '2'
        #  - string: 'none', 'small', 'complete' (case insensitive)
        try:
            mode = int(mode)
        except ValueError:
            mode_decoder = {'none': 0, 'small': 1, 'complete': 2}
            try:
                mode = mode_decoder[mode.lower()]
            except:
                mode = 0

        if mode not in (1, 2):
            return im

        if mode == 1:
            hsize = 32
            vsize = 24
        else:
            hsize = 64
            vsize = 54

        w, h = im.size

        with_border = Image.new("RGB", (w+2*hsize, h+2*vsize), color)
        with_border.paste(im, (hsize, vsize))
        return with_border

    return [_add_border(im, mode) for im in images]


def scale(images, factor):
    def _scale(im):
        if factor == 1:
            return im

        w, h = im.size
        return im.resize((int(w*factor), int(h*factor)), Image.ANTIALIAS)

    return [_scale(im) for im in images]

def interlace(images, alpha, scale):
    def _interlace(im):
        if scale not in (2, 3):
            return im
        if not 0 <= alpha <= 1:
            return im

        interlaced = im.copy()
        draw = ImageDraw.Draw(interlaced)

        w, h = im.size
        if scale == 2:
            for y in range(h / int(scale)):
                draw.line((0, 2*y, w, 2*y), fill=(0,0,0))
        else:
            for y in range(h / int(scale)):
                draw.line((0, 3*y, w, 3*y), fill=(0,0,0))
                draw.line((0, 3*y+2, w, 3*y+2), fill=(0,0,0))
        del draw

        return Image.blend(im, interlaced, alpha)

    return [_interlace(im) for im in images]

def light_blur(images):
    return [im.filter(LightBlur) for im in images]


class Picture(object):
    def __init__(self, *args, **kwargs):
        self.images = args
        self.border_color = kwargs.get('border_color', None)
        self.durations = kwargs.get('durations', None)
        self.format = kwargs.get('format', '<Unknown>')

    def postprocess(self, options_=None):
        options = {
            'scale': 1,
            'border-color': (0, 0, 0),
            'border-width': 'none',
            'interlace': 1,
            'blur': False
        }
        if options_:
            options.update(options_)

        images = self.images[:]

        color = self.border_color if self.border_color else options['border-color']
        images = add_border(images, options['border-width'], color)

        images = scale(images, options['scale'])
        images = interlace(images, options['interlace'], options['scale'])

        if options['blur']:
            images = light_blur(images)

        return images

    def save(self, dest, namebase, format='png', options=None):
        images = self.postprocess(options)

        if format == 'gif':
            images = [im.convert('P', dither=Image.NONE, palette=Image.ADAPTIVE) for im in images]

        if len(images) == 1:
            images[0].save(dest/"%s.%s" % (namebase, format))
        else:
            if self.durations is None or format != 'gif':
                if format != 'gif' and self.durations == (50, 50):
                    # save only the first frame of flashing screens
                    # if output format does not support animation
                    images[0].save(dest/"%s.%s" % (namebase, format))
                else:
                    for i, im in enumerate(images):
                        images[0].save(dest/"%s.%03d.%s" % (namebase, i, format))
            else:
                assert len(images) == len(self.durations)

                # ensure that all images have the same size
                base_size = images[0].size

                _images = []
                _durations = []

                for im, duration in zip(images, self.durations):
                    if im.size == base_size:
                        _images.append(im)
                        _durations.append(duration)

                write_animated_gif(file(dest/'%s.gif' % namebase, 'wb'), _images, _durations)
