import re

HEADER = 0x00
BODY = 0xFF

def i16(c):
    return ord(c[0]) + (ord(c[1])<<8)

_NTFS_INVALID_CHARS = r'?*/\:;<>'

_devices = [re.compile(re_, re.IGNORECASE) for re_ in ['com\d', 'lpt\d', 'prn\d', 'null']]

class TapeHeader(object):
    def __init__(self, type_, name, size, p1, p2):
        self.type = type_
        self.name = name
        self.size = size
        self.p1 = p1
        self.p2 = p2

    def _pc_name(self):
        name = self.name

        # remove leading and trailing whitespaces
        name = re.sub('^\s+', '', name)
        name = re.sub('\s+$', '', name)

        # handle device names
        for device_re in _devices:
            if device_re.match(name):
                name += '_'

        # replace invalid chars
        name = re.sub(r'[%s\s]' % _NTFS_INVALID_CHARS, '_', name)

        # we need some name in any case...
        if name == '':
            name = '_'

        return name

    pc_name = property(_pc_name, None, None)

def load(f):
    if hasattr(f, 'open'):
        fd = f.open('rb')
    elif isinstance(f, basestring):
        fd = open(f, 'rb')
    else:
        fd = f

    files = []
    header = None
    while True:
        data = fd.read(2)
        if len(data) == 0:
            break

        if len(data) != 2:
            # unexpected EOF
            return ()

        block_size = i16(data)

        content = fd.read(block_size)
        if len(content) != block_size:
            # unexpected EOF
            return ()

        block_type = ord(content[0])
        if block_type not in (HEADER, BODY):
            # unknown block type
            return ()

        if block_type == HEADER and block_size != 19:
            # wrong header size
            return ()

        checksum = 0
        for ch in content:
            checksum ^= ord(ch)

        if checksum != 0:
            # wrong block checksum
            return ()

        block = content[1:-1]
        if block_type == HEADER:
            hdr = TapeHeader(ord(block[0]), block[1:11], i16(block[11:13]),
                             i16(block[13:15]), i16(block[15:17]))
            if header:
                # two consecutive headers
                return ()
            header = hdr
        else:
            if header:
                if header.size != len(block):
                    return ()
                files.append((header, block))
                header = None
            else:
                # headerless block
                files.append((None, block))

    return files
