import re

from path import path
from collections import defaultdict
from ctypes import string_at, addressof, sizeof, cast, POINTER, create_string_buffer, memmove

from headers import HobetaHdr, SCLFileHdr, TRDFileHdr

class InvalidHobetaError(Exception):
    pass

_NTFS_INVALID_CHARS = r'?*/\:;<>'

def unique_names(hobeta_files, attr='hobeta_name'):
    result = []
    names = defaultdict(int)
    for hobeta in hobeta_files:
        fn = path(getattr(hobeta, attr))
        ln = fn.lower()
        if names[ln] > 0:
            fn = path('%s.%d%s' % (fn.namebase, names[ln], fn.ext))
        names[ln] += 1
        result.append(fn)
    return result


class HobetaFile(object):
    @staticmethod
    def checksum(data):
        if isinstance(data, HobetaHdr):
            data = string_at(addressof(data), sizeof(HobetaHdr))

        sum = 0
        for ch in data[:sizeof(HobetaHdr)-2]:
            sum = (sum + ord(ch)) & 0xffff
        sum = 257*sum + 105

        return sum & 0xffff

    @staticmethod
    def is_hobeta(content, size=None):
        if size is None:
            size = len(content)
        if size < sizeof(HobetaHdr) + 256:
            return False
        hdr = cast(content, POINTER(HobetaHdr))
        if size < sizeof(HobetaHdr) + hdr.contents.n_secs*256:
            return False

        # calculate HoBeta checksum
        sum = HobetaFile.checksum(content)
        return True if hdr.contents.checksum == sum else False

    @staticmethod
    def load(f):
        if getattr(f, 'read', None):
            return HobetaFile(f.read())
        if getattr(f, 'bytes', None):
            return HobetaFile(f.bytes())

        f = path(f)
        return HobetaFile(f.bytes())

    def __init__(self, data, header=None):
        if not header:
            if not HobetaFile.is_hobeta(data):
                raise InvalidHobetaError()
            self.header = HobetaHdr()
            buf = create_string_buffer(sizeof(HobetaHdr))
            memmove(addressof(self.header), data, sizeof(HobetaHdr))
            self.data = data[sizeof(HobetaHdr):]
        else:
            if isinstance(header, HobetaHdr):
                self.header = header
                self.data = data
            elif isinstance(header, (SCLFileHdr, TRDFileHdr)):
                self.header = HobetaHdr()
                for attr in ['name', 'type', 'start', 'size', 'n_secs']:
                    setattr(self.header, attr, getattr(header, attr))
                self.header.checksum = HobetaFile.checksum(self.header)
                self.data = data

    def __getattr__(self, attr):
        if attr in ['name', 'type', 'start', 'size', 'n_secs']:
            return getattr(self.header, attr)
        raise AttributeError, attr

    def _ext(self):
        if self.header.type == 'B' or self.header.ext in ['Cxi']:
            return self.header.type
        ext_re = re.compile(r"\d\d\d|[A-Za-z]\d\d|[A-Za-z]{2,2}\d|[A-Za-z]{3,3}")
        if ext_re.match(self.header.ext):
            return self.header.ext
        else:
            return self.header.type

    def _content(self):
        n_secs = self.header.size // 256
        if self.header.size % 256:
            n_secs += 1

        if n_secs == self.header.n_secs:
            return self.data[:self.header.size]
        else:
            return self.data

    def _name(self):
        name = self.header.name
        # handle device names
        for device_re in ['com\d', 'lpt\d', 'prn\d', 'null']:
            device_re = re.compile(device_re, re.IGNORECASE)
            if device_re.match(name):
                name += '_'

        # remove leading and trailing whitespaces
        name = re.sub('^\s+', '', name)
        name = re.sub('\s+$', '', name)

        # replace invalid chars
        name = re.sub(r'[%s\s]' % _NTFS_INVALID_CHARS, '_', name)

        # we need some name in any case...
        if name == '':
            name = '_'

        return name

    def _pc_name(self):
        ext = self.ext

        if len(ext) == 1:
            ext = re.sub(r'[%s\s]' % _NTFS_INVALID_CHARS, '_', ext)

        return "%s.%s" % (self._name(), ext)

    def _hobeta_name(self):
        ext = re.sub(r'[%s\s]' % _NTFS_INVALID_CHARS, '_', self.type)

        return "%s.$%s" % (self._name(), ext)

    def _trdos_name(self):
        ext = self.ext
        if len(ext) == 1:
            ext = '<%s>' % ext

        return "%s %s" % (self.name, ext)


    ext = property(_ext, None, None)
    content = property(_content, None, None)
    pc_name = property(_pc_name, None, None)
    hobeta_name = property(_hobeta_name, None, None)
    trdos_name = property(_trdos_name, None, None)
    deleted = property(lambda s: True if s.header.name.startswith(chr(1)) else False, None, None)
    hobeta = property(lambda s: string_at(addressof(s.header), sizeof(HobetaHdr))+s.data, None, None)


